/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.hyracks.algebricks.runtime.operators.win;

import java.nio.ByteBuffer;

import org.apache.hyracks.algebricks.data.IBinaryIntegerInspector;
import org.apache.hyracks.algebricks.data.IBinaryIntegerInspectorFactory;
import org.apache.hyracks.algebricks.runtime.base.IRunningAggregateEvaluatorFactory;
import org.apache.hyracks.algebricks.runtime.base.IScalarEvaluator;
import org.apache.hyracks.algebricks.runtime.base.IScalarEvaluatorFactory;
import org.apache.hyracks.api.comm.IFrame;
import org.apache.hyracks.api.comm.VSizeFrame;
import org.apache.hyracks.api.context.IHyracksTaskContext;
import org.apache.hyracks.api.dataflow.value.IBinaryComparator;
import org.apache.hyracks.api.dataflow.value.IBinaryComparatorFactory;
import org.apache.hyracks.api.exceptions.HyracksDataException;
import org.apache.hyracks.data.std.api.IPointable;
import org.apache.hyracks.data.std.primitive.VoidPointable;
import org.apache.hyracks.data.std.util.DataUtils;
import org.apache.hyracks.dataflow.common.comm.io.ArrayTupleBuilder;
import org.apache.hyracks.dataflow.common.comm.io.FrameTupleAccessor;
import org.apache.hyracks.dataflow.common.comm.util.FrameUtils;
import org.apache.hyracks.dataflow.common.data.accessors.FrameTupleReference;
import org.apache.hyracks.dataflow.common.data.accessors.IFrameTupleReference;
import org.apache.hyracks.dataflow.common.data.accessors.PointableTupleReference;
import org.apache.hyracks.dataflow.common.io.GeneratedRunFileReader;
import org.apache.hyracks.dataflow.std.group.IAggregatorDescriptor;
import org.apache.hyracks.storage.common.MultiComparator;

/**
 * Runtime for window operators that performs partition materialization and can evaluate running aggregates
 * as well as regular aggregates (in nested plans) over window frames.
 */
public class WindowNestedPlansPushRuntime extends WindowMaterializingPushRuntime {

    private final boolean frameValueExists;

    private final IScalarEvaluatorFactory[] frameValueEvalFactories;

    private IScalarEvaluator[] frameValueEvals;

    private PointableTupleReference frameValuePointables;

    private final IBinaryComparatorFactory[] frameValueComparatorFactories;

    private MultiComparator frameValueComparators;

    private final boolean frameStartExists;

    private final IScalarEvaluatorFactory[] frameStartEvalFactories;

    private IScalarEvaluator[] frameStartEvals;

    private PointableTupleReference frameStartPointables;

    private final boolean frameStartIsMonotonic;

    private final boolean frameEndExists;

    private final IScalarEvaluatorFactory[] frameEndEvalFactories;

    private IScalarEvaluator[] frameEndEvals;

    private PointableTupleReference frameEndPointables;

    private final boolean frameExcludeExists;

    private final IScalarEvaluatorFactory[] frameExcludeEvalFactories;

    private IScalarEvaluator[] frameExcludeEvals;

    private final int frameExcludeNegationStartIdx;

    private PointableTupleReference frameExcludePointables;

    private IPointable frameExcludePointable2;

    private final IBinaryComparatorFactory[] frameExcludeComparatorFactories;

    private IBinaryComparator[] frameExcludeComparators;

    private final boolean frameOffsetExists;

    private final IScalarEvaluatorFactory frameOffsetEvalFactory;

    private IScalarEvaluator frameOffsetEval;

    private IPointable frameOffsetPointable;

    private final IBinaryIntegerInspectorFactory binaryIntegerInspectorFactory;

    private final int frameMaxObjects;

    private final int nestedAggOutSchemaSize;

    private final WindowAggregatorDescriptorFactory nestedAggFactory;

    private IAggregatorDescriptor nestedAgg;

    private IFrame copyFrame2;

    private IFrame runFrame;

    private int runFrameChunkId;

    private long runFrameSize;

    private FrameTupleAccessor tAccess2;

    private FrameTupleReference tRef2;

    private IBinaryIntegerInspector bii;

    private int chunkIdxFrameStartGlobal;

    private int tBeginIdxFrameStartGlobal;

    private long readerPosFrameStartGlobal;

    WindowNestedPlansPushRuntime(int[] partitionColumns, IBinaryComparatorFactory[] partitionComparatorFactories,
            IBinaryComparatorFactory[] orderComparatorFactories, IScalarEvaluatorFactory[] frameValueEvalFactories,
            IBinaryComparatorFactory[] frameValueComparatorFactories, IScalarEvaluatorFactory[] frameStartEvalFactories,
            boolean frameStartIsMonotonic, IScalarEvaluatorFactory[] frameEndEvalFactories,
            IScalarEvaluatorFactory[] frameExcludeEvalFactories, int frameExcludeNegationStartIdx,
            IBinaryComparatorFactory[] frameExcludeComparatorFactories, IScalarEvaluatorFactory frameOffsetEvalFactory,
            IBinaryIntegerInspectorFactory binaryIntegerInspectorFactory, int frameMaxObjects, int[] projectionColumns,
            int[] runningAggOutColumns, IRunningAggregateEvaluatorFactory[] runningAggFactories,
            int nestedAggOutSchemaSize, WindowAggregatorDescriptorFactory nestedAggFactory, IHyracksTaskContext ctx) {
        super(partitionColumns, partitionComparatorFactories, orderComparatorFactories, projectionColumns,
                runningAggOutColumns, runningAggFactories, ctx);
        this.frameValueEvalFactories = frameValueEvalFactories;
        this.frameValueExists = frameValueEvalFactories != null && frameValueEvalFactories.length > 0;
        this.frameStartEvalFactories = frameStartEvalFactories;
        this.frameStartExists = frameStartEvalFactories != null && frameStartEvalFactories.length > 0;
        this.frameStartIsMonotonic = frameStartExists && frameStartIsMonotonic;
        this.frameEndEvalFactories = frameEndEvalFactories;
        this.frameEndExists = frameEndEvalFactories != null && frameEndEvalFactories.length > 0;
        this.frameValueComparatorFactories = frameValueComparatorFactories;
        this.frameExcludeEvalFactories = frameExcludeEvalFactories;
        this.frameExcludeExists = frameExcludeEvalFactories != null && frameExcludeEvalFactories.length > 0;
        this.frameExcludeComparatorFactories = frameExcludeComparatorFactories;
        this.frameExcludeNegationStartIdx = frameExcludeNegationStartIdx;
        this.frameOffsetExists = frameOffsetEvalFactory != null;
        this.frameOffsetEvalFactory = frameOffsetEvalFactory;
        this.binaryIntegerInspectorFactory = binaryIntegerInspectorFactory;
        this.frameMaxObjects = frameMaxObjects;
        this.nestedAggFactory = nestedAggFactory;
        this.nestedAggOutSchemaSize = nestedAggOutSchemaSize;
    }

    @Override
    protected void init() throws HyracksDataException {
        super.init();

        if (frameValueExists) {
            frameValueEvals = createEvaluators(frameValueEvalFactories, ctx);
            frameValueComparators = MultiComparator.create(frameValueComparatorFactories);
            frameValuePointables = createPointables(frameValueEvalFactories.length);
        }
        if (frameStartExists) {
            frameStartEvals = createEvaluators(frameStartEvalFactories, ctx);
            frameStartPointables = createPointables(frameStartEvalFactories.length);
        }
        if (frameEndExists) {
            frameEndEvals = createEvaluators(frameEndEvalFactories, ctx);
            frameEndPointables = createPointables(frameEndEvalFactories.length);
        }
        if (frameExcludeExists) {
            frameExcludeEvals = createEvaluators(frameExcludeEvalFactories, ctx);
            frameExcludeComparators = createBinaryComparators(frameExcludeComparatorFactories);
            frameExcludePointables = createPointables(frameExcludeEvalFactories.length);
            frameExcludePointable2 = VoidPointable.FACTORY.createPointable();
        }
        if (frameOffsetExists) {
            frameOffsetEval = frameOffsetEvalFactory.createScalarEvaluator(ctx);
            frameOffsetPointable = VoidPointable.FACTORY.createPointable();
            bii = binaryIntegerInspectorFactory.createBinaryIntegerInspector(ctx);
        }

        nestedAgg = nestedAggFactory.createAggregator(ctx, null, null, null, null, null, -1);

        runFrame = new VSizeFrame(ctx);
        copyFrame2 = new VSizeFrame(ctx);
        tAccess2 = new FrameTupleAccessor(inputRecordDesc);
        tRef2 = new FrameTupleReference();
    }

    @Override
    protected void beginPartitionImpl() throws HyracksDataException {
        super.beginPartitionImpl();
        chunkIdxFrameStartGlobal = -1;
        tBeginIdxFrameStartGlobal = -1;
        readerPosFrameStartGlobal = -1;
        runFrameChunkId = -1;
    }

    @Override
    protected void producePartitionTuples(int chunkIdx, GeneratedRunFileReader reader) throws HyracksDataException {
        boolean frameStartForward = frameStartIsMonotonic && chunkIdxFrameStartGlobal >= 0;

        long readerPos = -1;
        int nChunks = getPartitionChunkCount();
        if (nChunks > 1) {
            readerPos = reader.position();
            if (chunkIdx == 0) {
                ByteBuffer curFrameBuffer = curFrame.getBuffer();
                int pos = curFrameBuffer.position();
                copyFrame2.ensureFrameSize(curFrameBuffer.capacity());
                FrameUtils.copyAndFlip(curFrameBuffer, copyFrame2.getBuffer());
                curFrameBuffer.position(pos);
            }
        }

        tAccess.reset(curFrame.getBuffer());
        int tBeginIdx = getTupleBeginIdx(chunkIdx);
        int tEndIdx = getTupleEndIdx(chunkIdx);
        for (int tIdx = tBeginIdx; tIdx <= tEndIdx; tIdx++) {
            tRef.reset(tAccess, tIdx);

            // running aggregates
            produceTuple(tupleBuilder, tAccess, tIdx, tRef);

            // frame boundaries
            if (frameStartExists) {
                evaluate(frameStartEvals, tRef, frameStartPointables);
            }
            if (frameEndExists) {
                evaluate(frameEndEvals, tRef, frameEndPointables);
            }
            if (frameExcludeExists) {
                evaluate(frameExcludeEvals, tRef, frameExcludePointables);
            }
            int toSkip = 0;
            if (frameOffsetExists) {
                frameOffsetEval.evaluate(tRef, frameOffsetPointable);
                toSkip = bii.getIntegerValue(frameOffsetPointable.getByteArray(), frameOffsetPointable.getStartOffset(),
                        frameOffsetPointable.getLength());
            }
            int toWrite = frameMaxObjects;

            // aggregator created by WindowAggregatorDescriptorFactory does not process argument tuple in init()
            nestedAgg.init(null, null, -1, null);

            int chunkIdxInnerStart = frameStartForward ? chunkIdxFrameStartGlobal : 0;
            int tBeginIdxInnerStart = frameStartForward ? tBeginIdxFrameStartGlobal : -1;
            if (nChunks > 1) {
                reader.seek(frameStartForward ? readerPosFrameStartGlobal : 0);
            }

            int chunkIdxFrameStartLocal = -1, tBeginIdxFrameStartLocal = -1;
            long readerPosFrameStartLocal = -1;

            frame_loop: for (int chunkIdxInner = chunkIdxInnerStart; chunkIdxInner < nChunks; chunkIdxInner++) {
                long readerPosFrameInner;
                IFrame frameInner;
                if (chunkIdxInner == 0) {
                    // first chunk's frame is always in memory
                    frameInner = chunkIdx == 0 ? curFrame : copyFrame2;
                    readerPosFrameInner = 0;
                } else {
                    readerPosFrameInner = reader.position();
                    if (runFrameChunkId == chunkIdxInner) {
                        // runFrame has this chunk, so just advance the reader
                        reader.seek(readerPosFrameInner + runFrameSize);
                    } else {
                        reader.nextFrame(runFrame);
                        runFrameSize = reader.position() - readerPosFrameInner;
                        runFrameChunkId = chunkIdxInner;
                    }
                    frameInner = runFrame;
                }
                tAccess2.reset(frameInner.getBuffer());

                int tBeginIdxInner;
                if (tBeginIdxInnerStart < 0) {
                    tBeginIdxInner = getTupleBeginIdx(chunkIdxInner);
                } else {
                    tBeginIdxInner = tBeginIdxInnerStart;
                    tBeginIdxInnerStart = -1;
                }
                int tEndIdxInner = getTupleEndIdx(chunkIdxInner);

                for (int tIdxInner = tBeginIdxInner; tIdxInner <= tEndIdxInner; tIdxInner++) {
                    tRef2.reset(tAccess2, tIdxInner);

                    if (frameStartExists || frameEndExists) {
                        evaluate(frameValueEvals, tRef2, frameValuePointables);
                        if (frameStartExists) {
                            if (frameValueComparators.compare(frameValuePointables, frameStartPointables) < 0) {
                                // skip if value < start
                                continue;
                            }
                            if (chunkIdxFrameStartLocal < 0) {
                                // save position of the first tuple that matches the frame start.
                                // we'll continue from it in the next frame iteration
                                chunkIdxFrameStartLocal = chunkIdxInner;
                                tBeginIdxFrameStartLocal = tIdxInner;
                                readerPosFrameStartLocal = readerPosFrameInner;
                            }
                        }
                        if (frameEndExists
                                && frameValueComparators.compare(frameValuePointables, frameEndPointables) > 0) {
                            // skip and exit if value > end
                            break frame_loop;
                        }
                    }
                    if (frameExcludeExists && isExcluded()) {
                        // skip if excluded
                        continue;
                    }

                    if (toSkip > 0) {
                        // skip if offset hasn't been reached
                        toSkip--;
                        continue;
                    }

                    if (toWrite != 0) {
                        nestedAgg.aggregate(tAccess2, tIdxInner, null, -1, null);
                    }
                    if (toWrite > 0) {
                        toWrite--;
                    }
                    if (toWrite == 0) {
                        break frame_loop;
                    }
                }
            }

            nestedAgg.outputFinalResult(tupleBuilder, null, -1, null);
            appendToFrameFromTupleBuilder(tupleBuilder);

            if (frameStartIsMonotonic) {
                if (chunkIdxFrameStartLocal >= 0) {
                    chunkIdxFrameStartGlobal = chunkIdxFrameStartLocal;
                    tBeginIdxFrameStartGlobal = tBeginIdxFrameStartLocal;
                    readerPosFrameStartGlobal = readerPosFrameStartLocal;
                } else {
                    // frame start not found, set start beyond the last chunk
                    chunkIdxFrameStartGlobal = nChunks;
                    tBeginIdxFrameStartGlobal = 0;
                    readerPosFrameStartGlobal = 0;
                }
            }
        }

        if (nChunks > 1) {
            reader.seek(readerPos);
        }
    }

    private boolean isExcluded() throws HyracksDataException {
        for (int i = 0; i < frameExcludeEvals.length; i++) {
            frameExcludeEvals[i].evaluate(tRef2, frameExcludePointable2);
            boolean b = DataUtils.compare(frameExcludePointables.getField(i), frameExcludePointable2,
                    frameExcludeComparators[i]) != 0;
            if (i >= frameExcludeNegationStartIdx) {
                b = !b;
            }
            if (b) {
                return false;
            }
        }
        return true;
    }

    @Override
    protected ArrayTupleBuilder createOutputTupleBuilder(int[] projectionList) {
        return new ArrayTupleBuilder(projectionList.length + nestedAggOutSchemaSize);
    }

    private static IScalarEvaluator[] createEvaluators(IScalarEvaluatorFactory[] evalFactories, IHyracksTaskContext ctx)
            throws HyracksDataException {
        IScalarEvaluator[] evals = new IScalarEvaluator[evalFactories.length];
        for (int i = 0; i < evalFactories.length; i++) {
            evals[i] = evalFactories[i].createScalarEvaluator(ctx);
        }
        return evals;
    }

    private static void evaluate(IScalarEvaluator[] evals, IFrameTupleReference inTuple,
            PointableTupleReference outTuple) throws HyracksDataException {
        for (int i = 0; i < evals.length; i++) {
            evals[i].evaluate(inTuple, outTuple.getField(i));
        }
    }

    private static PointableTupleReference createPointables(int ln) {
        IPointable[] pointables = new IPointable[ln];
        for (int i = 0; i < ln; i++) {
            pointables[i] = VoidPointable.FACTORY.createPointable();
        }
        return new PointableTupleReference(pointables);
    }
}